﻿// <copyright file="TestHost.cs" company="T4 Toolbox Team">
//  Copyright © T4 Toolbox Team. All Rights Reserved.
// </copyright>

namespace T4Toolbox.DirectiveProcessors
{
    using System;
    using System.CodeDom.Compiler;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using Microsoft.VisualStudio.TextTemplating;

    /// <summary>
    /// Test host for directive processors
    /// </summary>
    internal class TestHost : ITextTemplatingEngineHost
    {
        /// <summary>
        /// Stores full path name of the <see cref="TemplateFile"/>
        /// </summary>
        private string templateFile;
        private AppDomain appDomain;

        /// <summary>
        /// Gets list of standard assembly references
        /// </summary>
        public IList<string> StandardAssemblyReferences
        {
            get { return new List<string>(new string[] { typeof(System.CodeDom.Compiler.CompilerError).Assembly.Location }); }
        }

        /// <summary>
        /// Gets list of standard namespaces
        /// </summary>
        public IList<string> StandardImports
        {
            get { return new List<string>(); }
        }

        /// <summary>
        /// Gets or sets file path of the template file.
        /// </summary>
        public string TemplateFile
        {
            get { return this.templateFile; }
            set { this.templateFile = value; }
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="optionName">Host option name</param>
        /// <returns>Host option value</returns>
        public object GetHostOption(string optionName)
        {
            // Test host doesn't support options
            return null;
        }

        /// <summary>
        /// This method is not implemented.
        /// </summary>
        /// <param name="requestFileName">Name of the included file</param>
        /// <param name="content">Content of the included file</param>
        /// <param name="location">Full path name of the included file</param>
        /// <returns>True if included file was loaded</returns>
        public bool LoadIncludeText(string requestFileName, out string content, out string location)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="errors">Compiler errors</param>
        public void LogErrors(CompilerErrorCollection errors)
        {
            // This method is intentionally left blank.
            // Test host ignores errors.
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="content">Template content</param>
        /// <returns>Templating app domain</returns>
        public AppDomain ProvideTemplatingAppDomain(string content)
        {
            // Caching the appDomain. This behavior is required to test the DynamicAssembyReferenceProcessor
            if (this.appDomain == null)
            {
                this.appDomain = AppDomain.CreateDomain("TextTemplatingAppDomain");
            }
            return this.appDomain;
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="assemblyReference">Assembly reference</param>
        /// <returns>Full path name of the assembly</returns>
        public string ResolveAssemblyReference(string assemblyReference)
        {
            return ResolvePath(assemblyReference);
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="processorName">Directive processor name</param>
        /// <returns>Directive processor type</returns>
        public Type ResolveDirectiveProcessor(string processorName)
        {
            if (StringComparer.OrdinalIgnoreCase.Equals(processorName, typeof(VolatileAssemblyProcessor).FullName))
            {
                return typeof(VolatileAssemblyProcessor);
            }
            throw new InvalidOperationException();
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="directiveId">directive ID</param>
        /// <param name="processorName">Directive processor name</param>
        /// <param name="parameterName">Parameter name</param>
        /// <returns>Default parameter value</returns>
        public string ResolveParameterValue(string directiveId, string processorName, string parameterName)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="path">Relative file path</param>
        /// <returns>Absolute path</returns>
        public string ResolvePath(string path)
        {
            if (Path.IsPathRooted(path))
            {
                return path;
            }

            return Path.Combine(Path.GetDirectoryName(this.GetType().Assembly.Location), path);
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="extension">Output file extension</param>
        public void SetFileExtension(string extension)
        {
            // Empty
        }

        /// <summary>
        /// This method is not implemented
        /// </summary>
        /// <param name="encoding">Encoding of the output file</param>
        /// <param name="fromOutputDirective">true or false</param>
        public void SetOutputEncoding(Encoding encoding, bool fromOutputDirective)
        {
            // Empty
        }

        /// <summary>
        /// This method is here to imitate the identical one provided by the VSHost in order to test 
        /// the dynamic assembly reference directive processor.
        /// </summary>
        private void UnloadGenerationAppDomain()
        {
            if (this.appDomain != null)
            {
                AppDomain.Unload(this.appDomain);
                this.appDomain = null;
            }
        }
    }
}
