﻿// <copyright file="DynamicAssemblyReferenceProcessorTest.cs" company="T4 Toolbox Team">
//  Copyright © T4 Toolbox Team. All Rights Reserved.
// </copyright>

namespace T4Toolbox.DirectiveProcessors
{
    using System;
    using System.IO;
    using System.CodeDom.Compiler;
    using System.Collections.Generic;
    using System.Reflection;
    using System.Reflection.Emit;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Microsoft.VisualStudio.TextTemplating;

    [TestClass]
    public class VolatileAssemblyProcessorTest
    {
        [TestMethod]
        public void Test_ProcessDirective()
        {
            const string testTemplate =
                "<#@ volatileassembly name=\"..\\TestAssembly.dll\" processor=\"T4Toolbox.VolatileAssemblyProcessor\" #>" +
                "<#TestType tt = new TestType();#>" +
                "<#= tt.Compute(3,4) #>";

            // Create a test assembly with a type that computes the product of 2 numbers
            CreateTestAssembly(true);
            // Run the template that calls the method in the test assembly
            string output = GetTemplateOutput(testTemplate);
            // We expect the output to be 12 (the product of 3 and 4)
            Assert.AreEqual(output.Trim(), "12");

            // Change the test assembly to compute the sum instead of the product
            CreateTestAssembly(false);
            output = GetTemplateOutput(testTemplate);
            // We expect the output to be 7 (the sum of 3 and 4)
            Assert.AreEqual(output.Trim(), "7");
        }

        [TestMethod]
        public void Test_ProcessDirective_InvalidAssemblyName()
        {
            VolatileAssemblyProcessor processor = new VolatileAssemblyProcessor();
            
            // Initialize
            TestHost host = new TestHost();
            host.TemplateFile = "Test.tt";
            processor.Initialize(host);            

            // Start Processing Run
            CompilerErrorCollection errors = new CompilerErrorCollection();
            CodeDomProvider languageProvider = new Microsoft.CSharp.CSharpCodeProvider();
            processor.StartProcessingRun(languageProvider, string.Empty, errors);

            // Process directive
            string assemblyName = "..\\InvalidAssembly.dll";
            Dictionary<string, string> arguments = new Dictionary<string,string>();
            arguments["name"] = assemblyName;
            processor.ProcessDirective("volatileassembly", arguments);

            Assert.AreEqual(1, errors.Count, "Invalid assembly name should be reported");
            Assert.AreEqual(errors[0].FileName, host.TemplateFile, "Error should reference the template file");
            StringAssert.Contains(errors[0].ErrorText, assemblyName, "Error message should include the actual path specified");
        }

        private void CreateTestAssembly(bool multiply)
        {
            // Create a dynamic assembly with a TestType that has a Compute(int a,int b) method that returns
            // either a*b or a+b.
            AssemblyName name = new AssemblyName("TestAssembly");
            AssemblyBuilder assembly = AppDomain.CurrentDomain.DefineDynamicAssembly(name, AssemblyBuilderAccess.Save);
            ModuleBuilder module = assembly.DefineDynamicModule(name.Name, name.Name + ".dll");

            TypeBuilder type = module.DefineType("TestType", TypeAttributes.Public);

            MethodBuilder method = type.DefineMethod("Compute", MethodAttributes.Public, typeof(int), new Type[] { typeof(int), typeof(int) });
            var ilGenerator = method.GetILGenerator();
            ilGenerator.Emit(OpCodes.Ldarg_1);
            ilGenerator.Emit(OpCodes.Ldarg_2);
            if (multiply)
            {
                ilGenerator.Emit(OpCodes.Mul);
            }
            else
            {
                ilGenerator.Emit(OpCodes.Add);
            }

            ilGenerator.Emit(OpCodes.Ret);

            type.CreateType();
            assembly.Save(name.Name + ".dll");

            // Move the file up one directory. This is to get it out of the probing path for the templating
            // AppDomain assembly loader.
            string assemblyPath = Path.Combine("..", name.Name + ".dll");
            if (File.Exists(assemblyPath))
            {
                File.Delete(assemblyPath);
            }
            File.Move(name.Name + ".dll", assemblyPath);
        }

        private string GetTemplateOutput(string templateText)
        {
            Engine engine = new Engine();
            TestHost host = new TestHost();
            host.TemplateFile = "Test.tt";
            return engine.ProcessTemplate(templateText, host);
        }
    }
}