﻿// <copyright file="XsdProcessorTest.cs" company="T4 Toolbox Team">
//  Copyright © T4 Toolbox Team. All Rights Reserved.
// </copyright>

namespace T4Toolbox.DirectiveProcessors
{
    using System;
    using System.IO;
    using System.CodeDom.Compiler;
    using System.Collections.Generic;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Microsoft.VisualStudio.TextTemplating;

    /// <summary>
    /// This is a test class for <see cref="XsdProcessor"/>
    /// </summary>
    [TestClass]
    public class XsdProcessorTest
    {
        /// <summary>
        /// This schema imports <see cref="NestedSchema"/> and defines an element a complex type defined in it.
        /// It contains <see cref="string.Format"/> placeholder in schemaLocation attribute of xs:import element.
        /// </summary>
        private const string ImportingSchema =
            @"<?xml version='1.0' encoding='utf-8'?>" +
            @"<xs:schema elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema' xmlns:ns='http://tempuri.org/NestedSchema'>" +
            @"  <xs:import namespace='http://tempuri.org/NestedSchema' schemaLocation='{0}' />" +
            @"  <xs:element name='TestClass' nillable='true' type='ns:TestClass' />" +
            @"</xs:schema>";

        /// <summary>
        /// This schema includes <see cref="StandaloneSchema"/>.
        /// It contains <see cref="string.Format"/> placeholder in schemaLocation attribute of xs:include element.
        /// </summary>
        private const string IncludingSchema =
            @"<?xml version='1.0' encoding='utf-8'?>" +
            @"<xs:schema elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
            @"    <xs:include schemaLocation='{0}' />" +
            @"</xs:schema>";

        /// <summary>
        /// This schema defines a single complex type, used to define an element by the <see cref="ImportingSchema"/>.
        /// </summary>
        private const string NestedSchema =
            @"<?xml version='1.0' encoding='utf-8'?>" +
            @"<xs:schema elementFormDefault='qualified' targetNamespace='http://tempuri.org/NestedSchema' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
            @"  <xs:complexType name='TestClass'>" +
            @"    <xs:sequence>" +
            @"      <xs:element minOccurs='1' maxOccurs='1' name='TestProperty' type='xs:int' />" +
            @"    </xs:sequence>" +
            @"  </xs:complexType>" +
            @"</xs:schema>";

        /// <summary>
        /// This is a standalone schema that defines both a complex type and an element that uses it.
        /// </summary>
        private const string StandaloneSchema =
            @"<?xml version='1.0' encoding='utf-8'?>" +
            @"<xs:schema elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
            @"  <xs:element name='TestClass' nillable='true' type='TestClass' />" +
            @"  <xs:complexType name='TestClass'>" +
            @"    <xs:sequence>" +
            @"      <xs:element minOccurs='1' maxOccurs='1' name='TestProperty' type='xs:int' />" +
            @"    </xs:sequence>" +
            @"  </xs:complexType>" +
            @"</xs:schema>";

        /// <summary>
        /// Errors and warnings produced by the directive <see cref="processor"/>
        /// </summary>
        private CompilerErrorCollection errors;
        
        /// <summary>
        /// Text templating host
        /// </summary>
        private TestHost host;

        /// <summary>
        /// XSD directive processor
        /// </summary>
        private XsdProcessor processor;

        /// <summary>
        /// List of temporary files created by during a test run.
        /// </summary>
        private List<string> temporaryFiles;

        /// <summary>
        /// Initializes environment before each test method
        /// </summary>
        [TestInitialize]
        public void TestInitialize()
        {
            this.temporaryFiles = new List<string>();

            // Create text templating host
            this.host = new TestHost();
            this.host.TemplateFile = this.CreateTemporaryFile();

            // Create directive processor
            this.processor = new XsdProcessor();
            this.processor.Initialize(this.host);

            // Prepare for directive processing
            CodeDomProvider languageProvider = new Microsoft.CSharp.CSharpCodeProvider();
            string templateContents = "<#@ xsd processor=\"XsdProcessor\" #>";
            this.errors = new CompilerErrorCollection();
            this.processor.StartProcessingRun(languageProvider, templateContents, this.errors);
        }

        /// <summary>
        /// Cleans up environment after each test method
        /// </summary>
        [TestCleanup]
        public void TestCleanup()
        {
            // Delete all temporary files
            foreach (string fileName in this.temporaryFiles)
            {
                if (File.Exists(fileName))
                {
                    File.Delete(fileName);
                }
            }
        }

        /// <summary>
        /// This test verifies import of ADO.NET Entity Framework schema
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_EntityFrameworkXsd()
        {
            string code = this.ProcessDirective(@"%VS90COMNTOOLS%\..\..\Xml\Schemas\Microsoft.Data.Entity.Design.Edmx.xsd");
            Assert.IsTrue(code.Contains("public partial class TEdmx"));
        }

        /// <summary>
        /// This test verifies import of LINQ to SQL schema
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_LinqToSqlXsd()
        {
            string code = this.ProcessDirective(@"%VS90COMNTOOLS%\..\..\Xml\Schemas\DbmlSchema.xsd");
            Assert.IsTrue(code.Contains("public partial class Database"));
        }

        /// <summary>
        /// This test verifies schema can be specified using an absolute file path.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_StandaloneSchemaWithAbsolutePath()
        {
            string schemaFile = this.CreateSchemaFile(StandaloneSchema);
            string code = this.ProcessDirective(schemaFile);
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// This test verifies that file path can contain environment variables.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_StandaloneSchemaWithEnvironmentPath()
        {
            string schemaFile = this.CreateSchemaFile(StandaloneSchema);
            string code = this.ProcessDirective(@"%TEMP%\" + Path.GetFileName(schemaFile));
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// This test verifies that schema can be specified as a file path relative to the location of the template.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_StandaloneSchemaWithRelativePath()
        {
            string schemaFile = this.CreateSchemaFile(StandaloneSchema);
            string code = this.ProcessDirective(Path.GetFileName(schemaFile));
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// This test verifies that absolute file paths used in SchemaLocation of xs:include element are supported.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_IncludedSchemaWithAbsolutePath()
        {
            string nestedSchemaFile = this.CreateSchemaFile(StandaloneSchema);
            string mainSchemaFile = this.CreateSchemaFile(IncludingSchema, nestedSchemaFile);
            string code = this.ProcessDirective(mainSchemaFile);
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// This test verifies that relative file paths used in SchemaLocation of xs:include element are supported.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_IncludedSchemaWithRelativePath()
        {
            string nestedSchemaFile = this.CreateSchemaFile(StandaloneSchema);
            string mainSchemaFile = this.CreateSchemaFile(IncludingSchema, Path.GetFileName(nestedSchemaFile));
            string code = this.ProcessDirective(mainSchemaFile);
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// This test verifies that <see cref="XsdProcessor.ProcessDirective"/> detects 
        /// and reports schema validation errors.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_InvalidSchema()
        {
            // Schema with invalid type (xs:invalidType)
            string invalidSchema = 
                @"<?xml version='1.0' encoding='utf-8'?>" +
                @"<xs:schema elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
                @"  <xs:element name='TestClass' nillable='true' type='TestClass' />" +
                @"  <xs:complexType name='TestClass'>" +
                @"    <xs:sequence>" +
                @"      <xs:element minOccurs='1' maxOccurs='1' name='TestProperty' type='xs:invalidType' />" +
                @"    </xs:sequence>" +
                @"  </xs:complexType>" +
                @"</xs:schema>";

            string invalidSchemaFile = this.CreateSchemaFile(invalidSchema);
            this.ProcessDirective(invalidSchemaFile);
            Assert.AreEqual(2, this.errors.Count, "XsdProcessor should report schema validation errors");
            Assert.AreEqual(invalidSchemaFile, this.errors[0].FileName, "XsdProcessor should have reported an error in the invalid schema");
            Assert.AreEqual(this.host.TemplateFile, this.errors[1].FileName, "XsdProcessor should have reported and error in the template file");
        }

        /// <summary>
        /// This test verifies that <see cref="XsdProcessor.ProcessDirective"/> detects 
        /// and reports schema validation errors.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_MalformedSchema()
        {
            string malformedSchema =
                @"<?xml version='1.0' encoding='utf-8'?>" +
                @"<xs:schema elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
                @"  <xs:element name='TestClass' nillable='true' type='TestClass' />" +
                @"  <xs:complexType name='TestClass'>" +
                @"    <xs:sequence>" +
                @"      <xs:element minOccurs='1' maxOccurs='1' name='TestProperty' type='xs:invalidType' />";
                //// The rest of the sechema is missing
                //// @"    </xs:sequence>" +
                //// @"  </xs:complexType>" +
                //// @"</xs:schema>";

            string malformedSchemaFile = this.CreateSchemaFile(malformedSchema);
            this.ProcessDirective(malformedSchemaFile);
            Assert.AreEqual(1, this.errors.Count, "XsdProcessor should report schema validation errors");
            Assert.AreEqual(malformedSchemaFile, this.errors[0].FileName, "XsdProcessor should have reported an error in the invalid schema");
        }

        /// <summary>
        /// This test verifies that absolute file paths used in SchemaLocation of xs:import element are supported.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_ImportedSchemaWithAbsolutePath()
        {
            string nestedSchemaFile = this.CreateSchemaFile(NestedSchema);
            string mainSchemaFile = this.CreateSchemaFile(ImportingSchema, nestedSchemaFile);
            string code = this.ProcessDirective(mainSchemaFile);
            this.VerifyTestClassCode(code);
        }
        
        /// <summary>
        /// This test verifies that relative file paths used in SchemaLocation of xs:import element are supported.
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_ImportedSchemaWithRelativePath()
        {
            string nestedSchemaFile = this.CreateSchemaFile(NestedSchema);
            string mainSchemaFile = this.CreateSchemaFile(ImportingSchema, Path.GetFileName(nestedSchemaFile));
            string code = this.ProcessDirective(mainSchemaFile);
            this.VerifyTestClassCode(code);
        }

        /// <summary>
        /// Verifies that <see cref="XsdProcessor.ProcessDirective"/> method generates
        /// a compiler error when File parameter is missing
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_MissingFileParameter()
        {
            Dictionary<string, string> arguments = new Dictionary<string, string>();
            this.processor.ProcessDirective("xsd", arguments);
            Assert.AreEqual(1, this.errors.Count, "XsdProcessor should report an error when file parameter is missing");
            Assert.AreEqual(this.host.TemplateFile, this.errors[0].FileName, "XsdProcessor should have reported error in the template file");
        }

        /// <summary>
        /// Verifies that <see cref="XsdProcessor.ProcessDirective"/> method generates
        /// a compiler error when imported schema file doesn't exist
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_NonExistentImportedFile()
        {
            string mainSchemaFile = this.CreateSchemaFile(ImportingSchema, "bogus.xsd");

            Dictionary<string, string> arguments = new Dictionary<string, string>();
            arguments["file"] = mainSchemaFile;
            this.processor.ProcessDirective("xsd", arguments);

            Assert.AreEqual(1, this.errors.Count, "XsdProcessor should report an error when schema imports non-existent file");
            Assert.AreEqual(mainSchemaFile, this.errors[0].FileName, "XsdProcessor should have reported error in the importing schema");
        }

        /// <summary>
        /// Verifies that <see cref="XsdProcessor.ProcessDirective"/> method generates
        /// a compiler error when included schema file doesn't exist
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_NonExistentIncludedFile()
        {
            string mainSchemaFile = this.CreateSchemaFile(IncludingSchema, "bogus.xsd");

            Dictionary<string, string> arguments = new Dictionary<string, string>();
            arguments["file"] = mainSchemaFile;
            this.processor.ProcessDirective("xsd", arguments);

            Assert.AreEqual(1, this.errors.Count, "XsdProcessor should report an error when schema includes non-existent file");
            Assert.AreEqual(mainSchemaFile, this.errors[0].FileName, "XsdProcessor should have reported error in the including schema");
        }

        /// <summary>
        /// Verifies that <see cref="XsdProcessor.ProcessDirective"/> method generates
        /// a compiler error when File parameter points to file that doesn't exist
        /// </summary>
        [TestMethod]
        public void Test_ProcessDirective_NonExistentSchemaFile()
        {
            Dictionary<string, string> arguments = new Dictionary<string, string>();
            arguments["file"] = "bogus.xsd";
            this.processor.ProcessDirective("xsd", arguments);
            Assert.AreEqual(1, this.errors.Count, "XsdProcessor should report an error when file points to non-existent file");
            Assert.AreEqual(this.host.TemplateFile, this.errors[0].FileName, "XsdProcessor should have reported error in the template file");
        }

        /// <summary>
        /// Verifies schema reported in bug 12947: http://t4toolbox.codeplex.com/WorkItem/View.aspx?WorkItemId=12947 
        /// </summary>
        [TestMethod]
        public void Test_Bug12947()
        {
            const string TestSchema = 
                @"<?xml version='1.0' encoding='utf-8'?>" +
                @"<xs:schema attributeFormDefault='unqualified' elementFormDefault='qualified' xmlns:xs='http://www.w3.org/2001/XMLSchema'>" +
                @"  <xs:element name='TestClass'>" +
                @"    <xs:complexType>" +
                @"      <xs:sequence>" +
                @"        <xs:element maxOccurs='unbounded' name='ArrayItem'>" +
                @"          <xs:simpleType>" +
                @"            <xs:restriction base='xs:string' />" +
                @"          </xs:simpleType>" +
                @"        </xs:element>" +
                @"      </xs:sequence>" +
                @"    </xs:complexType>" +
                @"  </xs:element>" +
                @"</xs:schema>";

            string schemaFile = this.CreateSchemaFile(TestSchema);
            string code = this.ProcessDirective(schemaFile);
            this.VerifyTestClassCode(code);
        }

        #region Private methods

        /// <summary>
        /// Creates a temporary schema file.
        /// </summary>
        /// <param name="contents">Schema contents</param>
        /// <returns>Name of the schema file</returns>
        private string CreateSchemaFile(string contents)
        {
            string fileName = this.CreateTemporaryFile();
            File.WriteAllText(fileName, contents);
            return fileName;
        }

        /// <summary>
        /// Creates a temporary schema file
        /// </summary>
        /// <param name="contentsFormat"><see cref="string.Format"/> string for the schema contents</param>
        /// <param name="args"><see cref="string.Format"/> arguments for the schema contents</param>
        /// <returns>Name of the schema file</returns>
        private string CreateSchemaFile(string contentsFormat, params object[] args)
        {
            return this.CreateSchemaFile(string.Format(contentsFormat, args));
        }

        /// <summary>
        /// Creates a temporary file
        /// </summary>
        /// <returns>Name of the temporary file</returns>
        private string CreateTemporaryFile()
        {
            string fileName = Path.GetTempFileName();
            fileName = fileName.ToUpperInvariant();
            this.temporaryFiles.Add(fileName);
            return fileName;
        }

        /// <summary>
        /// Simulates a directive processor run
        /// </summary>
        /// <param name="file">Value of the files directive parameter</param>
        /// <returns>Code generated by the directive processor</returns>
        private string ProcessDirective(string file)
        {
            Dictionary<string, string> arguments = new Dictionary<string, string>();
            arguments["file"] = file;
            this.processor.ProcessDirective("xsd", arguments);

            this.processor.FinishProcessingRun();

            return this.processor.GetClassCodeForProcessingRun();
        }

        /// <summary>
        /// Verifies that generated code contains a valid TestClass definition.
        /// </summary>
        /// <param name="code">C# source code</param>
        private void VerifyTestClassCode(string code)
        {
            Assert.IsTrue(code.Contains("public partial class TestClass"));
            Assert.IsTrue(code.Contains("public static TestClass Load(string filePath)"));
        }

        #endregion
    }
}
